import os
import shutil
import signal
import subprocess
import tempfile
import time
import types
import typing as t
import warnings
import weakref
from enum import Enum
from pathlib import Path

import httpx

from nats_tools.monitor import NATSMonitor
from nats_tools.templates import ConfigGenerator

DEFAULT_BIN_DIR = Path.home().joinpath("nats-server").absolute()


class InvalidWindowsSignal(Enum):
    SIGKILL = "KILL"
    SIGQUIT = "QUIT"
    SIGHUP = "HUP"
    SIGUSR1 = "USR1"
    SIGUSR2 = "USR2"


if os.name == "nt":

    class Signal(Enum):
        # Kill the process immediatley
        KILL = InvalidWindowsSignal.SIGKILL
        # Kills the process immediately and perform a core dump
        QUIT = InvalidWindowsSignal.SIGKILL
        # Stops the server grafefully
        STOP = signal.SIGTERM
        # Reopens the log file for log rotation
        REOPEN = InvalidWindowsSignal.SIGUSR1
        # Reload server configuration
        RELOAD = InvalidWindowsSignal.SIGHUP
        # Stops the server after evicting all clients
        LDM = InvalidWindowsSignal.SIGUSR2

else:

    class Signal(Enum):  # type: ignore[no-redef]
        # Kill the process immediatley
        KILL = signal.SIGKILL
        # Kills the process immediately and perform a core dump
        QUIT = signal.SIGQUIT
        # Stops the server grafefully
        STOP = signal.SIGTERM
        # Reopens the log file for log rotation
        REOPEN = signal.SIGUSR1
        # Reload server configuration
        RELOAD = signal.SIGHUP
        # Stops the server after evicting all clients
        LDM = signal.SIGUSR2


class NATSD:
    def __init__(
        self,
        address: str = "127.0.0.1",
        port: int = 4222,
        client_advertise: t.Optional[str] = None,
        server_name: t.Optional[str] = None,
        server_tags: t.Optional[t.Dict[str, str]] = None,
        user: t.Optional[str] = None,
        password: t.Optional[str] = None,
        users: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
        token: t.Optional[str] = None,
        http_port: int = 8222,
        debug: t.Optional[bool] = None,
        trace: t.Optional[bool] = None,
        trace_verbose: t.Optional[bool] = None,
        logtime: t.Optional[bool] = None,
        pid_file: t.Union[str, Path, None] = None,
        port_file_dir: t.Union[str, Path, None] = None,
        log_file: t.Union[str, Path, None] = None,
        log_size_limit: t.Optional[int] = None,
        tls_cert: t.Union[str, Path, None] = None,
        tls_key: t.Union[str, Path, None] = None,
        tls_ca_cert: t.Union[str, Path, None] = None,
        cluster_name: t.Optional[str] = None,
        cluster_url: t.Optional[str] = None,
        cluster_listen: t.Optional[str] = None,
        routes: t.Optional[t.List[str]] = None,
        no_advertise: t.Optional[bool] = None,
        with_jetstream: bool = False,
        jetstream_domain: t.Optional[str] = None,
        store_directory: t.Union[str, Path, None] = None,
        max_memory_store: t.Optional[int] = None,
        max_file_store: t.Optional[int] = None,
        max_outstanding_catchup: t.Optional[int] = None,
        allow_leafnodes: bool = False,
        leafnodes_listen_address: t.Optional[str] = None,
        leafnodes_listen_port: t.Optional[int] = None,
        leafnode_remotes: t.Optional[t.Dict[str, t.Any]] = None,
        websocket_listen_address: t.Optional[str] = None,
        websocket_listen_port: t.Optional[int] = None,
        websocket_advertise_url: t.Optional[str] = None,
        websocket_tls: t.Optional[bool] = None,
        websocket_tls_cert: t.Union[str, Path, None] = None,
        websocket_tls_key: t.Union[str, Path, None] = None,
        websocket_same_origin: t.Optional[bool] = None,
        websocket_allowed_origins: t.Optional[t.List[str]] = None,
        websocket_compression: t.Optional[bool] = None,
        jwt_path: t.Union[str, Path, None] = None,
        operator: t.Optional[str] = None,
        system_account: t.Optional[str] = None,
        system_account_jwt: t.Optional[str] = None,
        allow_delete_jwt: t.Optional[bool] = None,
        compare_jwt_interval: t.Optional[str] = None,
        resolver_preload: t.Optional[t.Dict[str, str]] = None,
        config_file: t.Union[str, Path, None] = None,
        max_cpus: t.Optional[float] = None,
        start_timeout: float = 1,
    ) -> None:
        """Create a new instance of nats-server daemon.

        Arguments:
            address: host address nats-server should listen to. Default is 127.0.0.1 (localhost).
            port: tcp port nats-server should listen to. Clients can connect to this port. Default is 4222.
            server_name: the server name. Default to auto-generated name.
            user: username required for connections. Omitted by default.
            password: password required for connections. Omitted by default.
            token: authorization token required for connections. Omitted by default.
            http_port: port for http monitoring. Default is 8222.
            debug: enable debugging output. Default is False.
            trace: enable raw traces. Default is False.
            pid_file: file to write process ID to. Omitted by default.
            log_file: file to redirect log output to. Omitted by default.
            tls_cert: server certificate file (TLS is enabled when both cert and key are provided)
            tls_key: server key file (TLS is enabled when both cert and key are provided)
            tls_ca_cert: client certificate for CA verification (mutual TLS is enabled when ca cert is provided)
            cluster_name: the cluster name. Default to auto-generated name when clustering is enabled.
            cluster_url: cluster URL for sollicited routes.
            cluster_listen: cluster URL from which members can solicite routes. Enable cluster mode when set.
            routes: routes to solicit and connect.
            no_advertise: do not advertise known cluster information to clients.
            with_jetstream: enable jetstream engine when True. Disabled by default.
            store_directory: path to jetstream store directory. Default to a temporary directory.
            config_file: path to a configuration file. None by default.
            max_cpus: maximum number of CPU configured using GOMAXPROCS environment variable. By default all CPUs can be used.
            start_timeout: amount of time to wait before raising an error when starting the daemon with wait=True.
        """
        if config_file is None:
            config_file = Path(tempfile.mkdtemp()).joinpath("nats.conf")
            generator = ConfigGenerator()
            config_str = generator.render(
                address=address,
                port=port,
                client_advertise=client_advertise,
                server_name=server_name,
                server_tags=server_tags,
                user=user,
                password=password,
                users=users,
                token=token,
                http_port=http_port,
                debug=debug,
                trace=trace,
                trace_verbose=trace_verbose,
                logtime=logtime,
                pid_file=pid_file,
                port_file_dir=port_file_dir,
                log_file=log_file,
                log_size_limit=log_size_limit,
                tls_cert=tls_cert,
                tls_key=tls_key,
                tls_ca_cert=tls_ca_cert,
                cluster_name=cluster_name,
                cluster_url=cluster_url,
                cluster_listen=cluster_listen,
                routes=routes,
                no_advertise=no_advertise,
                with_jetstream=with_jetstream,
                jetstream_domain=jetstream_domain,
                store_directory=store_directory,
                max_memory_store=max_memory_store,
                max_file_store=max_file_store,
                max_outstanding_catchup=max_outstanding_catchup,
                allow_leafnodes=allow_leafnodes,
                leafnodes_listen_address=leafnodes_listen_address,
                leafnodes_listen_port=leafnodes_listen_port,
                leafnode_remotes=leafnode_remotes,
                websocket_listen_address=websocket_listen_address,
                websocket_listen_port=websocket_listen_port,
                websocket_advertise_url=websocket_advertise_url,
                websocket_tls=websocket_tls,
                websocket_tls_cert=websocket_tls_cert,
                websocket_tls_key=websocket_tls_key,
                websocket_same_origin=websocket_same_origin,
                websocket_allowed_origins=websocket_allowed_origins,
                websocket_compression=websocket_compression,
                jwt_path=jwt_path,
                operator=operator,
                system_account=system_account,
                system_account_jwt=system_account_jwt,
                allow_delete_jwt=allow_delete_jwt,
                compare_jwt_interval=compare_jwt_interval,
                resolver_preload=resolver_preload,
            )
            config_file.write_text(config_str)
            weakref.finalize(self, shutil.rmtree, config_file.parent, True)
        self.server_name = server_name
        self.address = address
        self.port = port
        self.user = user
        self.password = password
        self.timeout = start_timeout
        self.http_port = http_port
        self.token = token
        self.bin_name = "nats-server"
        self.bin_path: t.Optional[str] = None
        self.config_file = Path(config_file) if config_file else None
        self.debug = debug or os.environ.get("DEBUG_NATS_TEST", "") in (
            "true",
            "1",
            "y",
            "yes",
            "on",
        )
        self.trace = trace or os.environ.get("DEBUG_NATS_TEST", "") in (
            "true",
            "1",
            "y",
            "yes",
            "on",
        )
        self.pid_file = Path(pid_file).absolute().as_posix() if pid_file else None
        self.log_file = Path(log_file).absolute().as_posix() if log_file else None
        self.max_cpus = max_cpus

        self.tls_cert = tls_cert
        self.tls_key = tls_key
        self.tls_ca_cert = tls_ca_cert
        if self.tls_ca_cert and self.tls_cert and self.tls_key:
            self.tls_verify = True
            self.tls = False
        elif self.tls_cert and self.tls_key:
            self.tls_verify = False
            self.tls = True
        elif self.tls_ca_cert:
            raise ValueError(
                "Both certificate and key files must be provided with a CA certificate"
            )
        elif self.tls_cert or self.tls_key:
            raise ValueError("Both certificate and key files must be provided")
        else:
            self.tls = False
            self.tls_verify = False

        self.cluster_name = cluster_name
        self.cluster_url = cluster_url
        self.cluster_listen = cluster_listen
        self.routes = routes
        self.no_advertise = no_advertise

        self.jetstream_enabled = with_jetstream
        if store_directory:
            self.store_dir = Path(store_directory)
            self._store_dir_is_temporary = False
        else:
            self.store_dir = Path(tempfile.mkdtemp()).resolve(True)
            self._store_dir_is_temporary = True
            weakref.finalize(self, shutil.rmtree, self.store_dir.as_posix(), True)

        self.proc: t.Optional["subprocess.Popen[bytes]"] = None
        self.monitor = NATSMonitor(f"http://{self.address}:{self.http_port}")

    def is_alive(self) -> bool:
        if self.proc is None:
            return False
        return self.proc.poll() is None

    def _cleanup_on_exit(self) -> None:
        if self.proc and self.proc.poll() is None:
            print(
                "[\033[0;31mWARNING\033[0;0m] Stopping server listening on %d."
                % self.port
            )
            self.kill()

    def start(self, wait: bool = False) -> "NATSD":
        # Check if there is an nats-server binary in the current working directory
        if Path(self.bin_name).is_file():
            self.bin_path = Path(self.bin_name).resolve(True).as_posix()
        # Path in `../scripts/install_nats.sh`
        elif DEFAULT_BIN_DIR.joinpath(self.bin_name).is_file():
            self.bin_path = DEFAULT_BIN_DIR.joinpath(self.bin_name).as_posix()
        # This directory contains binary
        else:
            self.bin_path = shutil.which(self.bin_name)
            if self.bin_path is None:
                raise FileNotFoundError("nats-server executable not found")

        cmd = [
            self.bin_path,
            "-p",
            "%d" % self.port,
            "-m",
            "%d" % self.http_port,
            "-a",
            self.address,
        ]

        if self.config_file is not None:
            if not self.config_file.exists():
                raise FileNotFoundError(self.config_file)
            else:
                config_file = self.config_file.absolute().as_posix()
            cmd.append("--config")
            cmd.append(config_file)

        env = os.environ.copy()

        if self.max_cpus:
            env["GOMAXPROCS"] = format(self.max_cpus, ".2f")

        if self.debug:
            self.proc = subprocess.Popen(cmd, env=env)
        else:
            self.proc = subprocess.Popen(
                cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, env=env
            )

        if self.debug:
            print(
                "[\033[0;33mDEBUG\033[0;0m] Server listening on port %d started."
                % self.port
            )
        if wait:
            deadline = time.time() + self.timeout or float("inf")
            while True:
                status = self.proc.poll()
                if status is not None:
                    if self.debug:
                        print(
                            "[\033[0;31mWARNING\033[0;0m] Server listening on port {port} already finished running with exit {ret}".format(
                                port=self.port, ret=self.proc.returncode
                            )
                        )
                    raise subprocess.CalledProcessError(
                        returncode=self.proc.returncode, cmd=self.proc.args
                    )
                if time.time() > deadline:
                    self.stop()
                    raise TimeoutError(
                        f"nats-server failed to start before timeout ({self.timeout:.3f}s)"
                    )
                try:
                    self.monitor.varz()
                    break
                except httpx.HTTPError as exc:
                    print(
                        f"[\033[0;31mDEBUG\033[0;0m] Waiting for server to be up. Last error: {type(exc).__name__} - {repr(exc)}."
                    )
                    time.sleep(0.1)
                    continue

        weakref.finalize(self, self._cleanup_on_exit)
        return self

    def stop(self, timeout: t.Optional[float] = 10) -> None:
        if self.debug:
            print(
                "[\033[0;33mDEBUG\033[0;0m] Server listening on %d will stop."
                % self.port
            )

        if self.proc is None:
            if self.debug:
                print(
                    "[\033[0;31mWARNING\033[0;0m] Failed terminating server listening on port %d"
                    % self.port
                )

        elif self.proc.returncode is not None:
            if self.debug:
                print(
                    "[\033[0;31mWARNING\033[0;0m] Server listening on port {port} already finished running with exit {ret}".format(
                        port=self.port, ret=self.proc.returncode
                    )
                )
        else:
            try:
                self.term(timeout=timeout)
            except TimeoutError:
                self.kill()
            if self.debug:
                print(
                    "[\033[0;33mDEBUG\033[0;0m] Server listening on %d was stopped."
                    % self.port
                )
        expected = 15 if os.name == "nt" else 1
        if self.proc and self.proc.returncode != expected:
            raise subprocess.CalledProcessError(
                self.proc.returncode, cmd=self.proc.args
            )

    def wait(self, timeout: t.Optional[float] = None) -> int:
        """Wait for process to finish and return status code.

        Possible status codes (non-exhaustive):
            -1: process is not started yet.
            0: process has been stopped after entering lame duck mode.
            1: process has been stopped due to TERM signal.
            2: process has been stopped due to QUIT signal.
            -9: process has been stopped due to KILL signal.
        """
        if self.proc is None:
            return 0
        status = self.proc.poll()
        if status is not None:
            return status
        return self.proc.wait(timeout=timeout)

    def send_signal(self, sig: t.Union[int, signal.Signals, Signal]) -> None:
        if self.proc is None:
            raise TypeError("Process is not started yet")
        status = self.proc.poll()
        if status is not None:
            raise subprocess.CalledProcessError(status, cmd=self.proc.args)
        if os.name != "nt":
            if not isinstance(sig, Signal):
                sig = signal.Signals(sig)
                sig = Signal(sig)
            os.kill(self.proc.pid, sig.value)
        else:
            sig = Signal(sig)
            if isinstance(sig.value, InvalidWindowsSignal):
                # Use a subprocess to explicitely call `nats-server --signal` which will handle signal correctly on Windows
                if sig.value == InvalidWindowsSignal.SIGKILL:
                    os.kill(self.proc.pid, signal.SIGINT)
                elif sig.value == InvalidWindowsSignal.SIGQUIT:
                    os.kill(self.proc.pid, signal.SIGBREAK)  # type: ignore[attr-defined]
                elif sig.value == InvalidWindowsSignal.SIGHUP:
                    warnings.warn("Config reload is not supported on Windows")
                elif sig.value == InvalidWindowsSignal.SIGUSR1:
                    warnings.warn("Log file roration is not supported on Windows")
                elif sig.value == InvalidWindowsSignal.SIGUSR2:
                    warnings.warn("Lame Duck Mode is not supported on Windows")
                    os.kill(self.proc.pid, signal.SIGINT)
            else:
                os.kill(self.proc.pid, sig.value)

    def quit(self, timeout: t.Optional[float] = None) -> None:
        self.send_signal(Signal.QUIT)
        self.wait(timeout=timeout)

    def kill(self, timeout: t.Optional[float] = None) -> None:
        self.send_signal(Signal.KILL)
        self.wait(timeout=timeout)

    def term(self, timeout: t.Optional[float] = 10) -> None:
        self.send_signal(Signal.STOP)
        self.wait(timeout=timeout)

    def reopen_log_file(self) -> None:
        self.send_signal(Signal.REOPEN)

    def enter_lame_duck_mode(self) -> None:
        self.send_signal(Signal.LDM)

    def reload_config(self) -> None:
        self.send_signal(Signal.RELOAD)

    def __enter__(self) -> "NATSD":
        return self.start(wait=True)

    def __exit__(
        self,
        error_type: t.Optional[t.Type[BaseException]] = None,
        error: t.Optional[BaseException] = None,
        traceback: t.Optional[types.TracebackType] = None,
    ) -> None:
        self.stop()



"""The `__about__` module exposes the version of the `nats_tools` package:

Example:

```python
from nats_tools.__about__ import __version__
print(__version__)
```

Note that container version can also be imported directly from `nats_tools` package:

Example:

```python
from nats_tools import __version__
print(__version__)
```
"""
__version__ = "1.0.1"



import types
import typing as t
from enum import Enum

import httpx


class SortOption(str, Enum):
    # Connection ID
    CID = "cid"
    # Connection start time, same as CID
    START = "start"
    # Number of subscriptions
    SUBS = "subs"
    # Amount of data in bytes waiting to be sent to client
    PENDING = "pending"
    # Number of messages sent
    MSGS_TO = "msgs_to"
    # Number of messages received
    MSGS_FROM = "msgs_from"
    # Number of bytes sent
    BYTES_TO = "bytes_to"
    # Number of bytes received
    BYTES_FROM = "bytes_from"
    # Last activity
    LAST = "last"
    # Amount of inactivity
    IDLE = "idle"
    # Lifetime of the connection
    UPTIME = "uptime"
    # Stop time for a closed connection
    STOP = "stop"
    # Reason for a closed connection
    REASON = "reason"


class SubsOption(Enum):
    FALSE = False
    TRUE = True
    DETAIL = "detail"


class StateOption(str, Enum):
    CLOSED = "closed"
    OPEN = "open"
    ANY = "any"


class NATSMonitor:
    def __init__(self, endpoint: str) -> None:
        self.endpoint = endpoint
        self._client: t.Optional[httpx.Client] = None

    def _request(self, endpoint: str, **params: t.Any) -> t.Dict[str, t.Any]:
        if self._client is None:
            self._client = httpx.Client(base_url=self.endpoint)
        response = self._client.get(endpoint, params=params)
        response.raise_for_status()
        return t.cast(t.Dict[str, t.Any], response.json())

    def varz(self) -> t.Dict[str, t.Any]:
        """The /varz endpoint returns general information about the server state and configuration.

        Example: https://demo.nats.io:8222/varz
        """
        return self._request("/varz")

    def jsz(
        self,
        acc: t.Optional[str] = None,
        accounts: t.Optional[bool] = None,
        streams: t.Optional[bool] = None,
        consumers: t.Optional[bool] = None,
        config: t.Optional[bool] = None,
        leader_only: bool = False,
        offset: int = 0,
        limit: int = 1024,
    ) -> t.Dict[str, t.Any]:
        """The /jsz endpoint reports more detailed information on JetStream.

        For accounts, it uses a paging mechanism that defaults to 1024 connections.

        NOTE: If you're in a clustered environment, it is recommended to retrieve the information
              from the stream's leader in order to get the most accurate and up-to-date data.

        Arguments:
            acc: include metrics for the specified account only. Omitted by default.
            accounts: include account specific jetstream information. Default is False.
            streams: include streams. When set, implies `accounts=True`. Default is False.
            consumers: include consumers. When set, implies `stream=True`. Default is False.
            config: when stream or consumer are requested, include their respective configuration. Default is False.
            leader_only: only the leader responds. Default is False.
            offset: pagination offset. Default is 0.
            limit: number of results to return. Default is 1024.

        Returns:
            results as a dictionary.
        """
        params: t.Dict[str, t.Any] = {
            "leader-only": leader_only,
            "limit": limit,
            "offset": offset,
        }
        if accounts:
            params["accounts"] = accounts
        if streams:
            params["streams"] = streams
        if consumers:
            params["consumers"] = consumers
        if config:
            params["config"] = config
        if acc:
            params["acc"] = acc
        return self._request("/jsz", **params)

    def connz(
        self,
        sort: t.Union[str, SortOption] = SortOption.CID,
        auth: bool = False,
        subs: t.Union[bool, str, SubsOption] = SubsOption.FALSE,
        offset: int = 0,
        limit: int = 1024,
        cid: t.Optional[int] = None,
        state: t.Union[str, StateOption] = StateOption.OPEN,
        mqtt_client: t.Optional[str] = None,
    ) -> t.Dict[str, t.Any]:
        """The /connz endpoint reports more detailed information on current and recently closed connections.

        It uses a paging mechanism which defaults to 1024 connections.

        Arguments:
            sort: sorts the results. Default is connection ID.
            auth: include username. Default is False.
            subs: include subscriptions. Default is False. When set to "detail", a list with more detailed subscription information is returned.
            offset: pagination offset. Default is 0.
            limit: number of results to return. Default is 1024.
            cid: return result for a single connection by its id. Omitted by default.
            state: return results for connections of particular state. Default is "open".
            mqtt_client: return results for connections with this MQTT client id. Omitted by default.

        Returns:
            results as a dictionary.
        """
        if not isinstance(sort, SortOption):
            sort = SortOption(sort)
        if not isinstance(subs, SubsOption):
            subs = SubsOption(subs)
        if not isinstance(state, StateOption):
            state = StateOption(state)
        params = {
            "sort": sort.value,
            "auth": 1 if auth else 0,
            "subs": subs.value,
            "offset": offset,
            "limit": limit,
        }
        if cid:
            params["cid"] = int(cid)
        if mqtt_client:
            params["mqtt_client"] = mqtt_client
        return self._request("/connz", **params)

    def accountz(self, acc: t.Optional[str] = None) -> t.Dict[str, t.Any]:
        """The /accountz endpoint reports information on a server's active accounts.

        The default behavior is to return a list of all accounts known to the server.

        Arguments:
            acc: include metrics for the specified account only. Default is empty. When not set
                a list of all accounts is included.
        """
        params: t.Dict[str, t.Any] = {}
        if acc:
            params["acc"] = acc
        return self._request("/accountz", **params)

    def accstatz(self, unused: bool = False) -> t.Dict[str, t.Any]:
        """The /accstatz endpoint reports per-account statistics such as the number of connections, messages/bytes in/out, etc.

        Arguments:
            unused: include accounts that do not have any current connections when True. Default is False.

        Returns:
            results as a dictionary.
        """
        return self._request("/accstatz", unused=unused)

    def subsz(
        self,
        subs: bool = False,
        offset: int = 0,
        limit: int = 1024,
        test: t.Optional[str] = None,
    ) -> t.Dict[str, t.Any]:
        """The /subsz endpoint reports detailed information about the current subscriptions and the routing data structure.
        It is not normally used.

        Arguments:
            subs: include subscriptions. Default is false.
            offset: pagination offset. Default is 0.
            limit: number of results to return. Default is 1024.
            test: test whether a subscription exists.

        Returns:
            results as a dictionary.
        """
        params: t.Dict[str, t.Any] = {
            "subs": subs,
            "offset": offset,
            "limit": limit,
        }
        if test:
            params["test"] = test
        return self._request("/subsz", **params)

    def routez(
        self, subs: t.Union[bool, str, SubsOption] = SubsOption.FALSE
    ) -> t.Dict[str, t.Any]:
        """The /routez endpoint reports information on active routes for a cluster.

        Routes are expected to be low, so there is no paging mechanism with this endpoint.

        Arguments:
            subs: include subscriptions. Default is False. When set to "detail", a list with more details subscription information is returned.

        Returns:
            results as a dictionary.
        """
        if not isinstance(subs, SubsOption):
            subs = SubsOption(subs)
        return self._request("/routez", subs=subs.value)

    def leafz(self, subs: bool = False) -> t.Dict[str, t.Any]:
        """The /leafz endpoint reports detailed information about the leaf node connections.

        Arguments:
            subs: include internal subscriptions. Default is False.

        Returns:
            results as dict
        """
        return self._request("/leafz", subs=subs)

    def gatewayz(
        self,
        accs: bool = False,
        gw_name: t.Optional[str] = None,
        acc_name: t.Optional[str] = None,
    ) -> t.Dict[str, t.Any]:
        """The /gatewayz endpoint reports information about gateways used to create a NATS supercluster.

        Like routes, the number of gateways are expected to be low, so there is no paging mechanism with this endpoint.

        Arguments:
            accs: include account information. Default is false.
            gw_name: return only remote gateways with this name. Omitted by default.
            acc_name: limit the list of accounts to this account name. Omitted by default.

        Returns:
            results as dict
        """
        params: t.Dict[str, t.Any] = {"accs": bool(accs)}
        if gw_name:
            params["gw_name"] = gw_name
        if acc_name:
            params["acc_name"] = acc_name
        return self._request("/gatewayz", **params)

    def healthz(
        self,
        js_enabled: bool = False,
        js_server_only: bool = False,
    ) -> t.Dict[str, t.Any]:
        """The /healthz endpoint returns OK if the server is able to accept connections.

        Arguments:
            js_enabled: returns an error if jetstream is disabled. Omitted by default.
            js_server_only: skip healthcheck of accounts, streams and consumers. Omitted by default.

        Returns:
            results as dictionary.
        """
        params: t.Dict[str, t.Any] = {}
        if js_enabled:
            params["js_enabled"] = 1
        if js_server_only:
            params["js_server_only"] = 1
        return self._request("/healthz", **params)


class AsyncNATSMonitor:
    def __init__(self, endpoint: str) -> None:
        self.endpoint = endpoint
        self._client: t.Optional[httpx.AsyncClient] = None

    async def _request(self, endpoint: str, **params: t.Any) -> t.Dict[str, t.Any]:
        if self._client is None:
            self._client = httpx.AsyncClient(base_url=self.endpoint)
        response = await self._client.get(endpoint, params=params)
        response.raise_for_status()
        return t.cast(t.Dict[str, t.Any], response.json())

    async def varz(self) -> t.Dict[str, t.Any]:
        """The /varz endpoint returns general information about the server state and configuration.

        Example: https://demo.nats.io:8222/varz
        """
        return await self._request("/varz")

    async def jsz(
        self,
        acc: t.Optional[str] = None,
        accounts: t.Optional[bool] = None,
        streams: t.Optional[bool] = None,
        consumers: t.Optional[bool] = None,
        config: t.Optional[bool] = None,
        leader_only: bool = False,
        offset: int = 0,
        limit: int = 1024,
    ) -> t.Dict[str, t.Any]:
        """The /jsz endpoint reports more detailed information on JetStream.

        For accounts, it uses a paging mechanism that defaults to 1024 connections.

        NOTE: If you're in a clustered environment, it is recommended to retrieve the information
              from the stream's leader in order to get the most accurate and up-to-date data.

        Arguments:
            acc: include metrics for the specified account only. Omitted by default.
            accounts: include account specific jetstream information. Default is False.
            streams: include streams. When set, implies `accounts=True`. Default is False.
            consumers: include consumers. When set, implies `stream=True`. Default is False.
            config: when stream or consumer are requested, include their respective configuration. Default is False.
            leader_only: only the leader responds. Default is False.
            offset: pagination offset. Default is 0.
            limit: number of results to return. Default is 1024.

        Returns:
            results as a dictionary.
        """
        params: t.Dict[str, t.Any] = {
            "leader-only": leader_only,
            "limit": limit,
            "offset": offset,
        }
        if accounts:
            params["accounts"] = accounts
        if streams:
            params["streams"] = streams
        if consumers:
            params["consumers"] = consumers
        if config:
            params["config"] = config
        if acc:
            params["acc"] = acc
        return await self._request("/jsz", **params)

    async def connz(
        self,
        sort: t.Union[str, SortOption] = SortOption.CID,
        auth: bool = False,
        subs: t.Union[bool, str, SubsOption] = SubsOption.FALSE,
        offset: int = 0,
        limit: int = 1024,
        cid: t.Optional[int] = None,
        state: t.Union[str, StateOption] = StateOption.OPEN,
        mqtt_client: t.Optional[str] = None,
    ) -> t.Dict[str, t.Any]:
        """The /connz endpoint reports more detailed information on current and recently closed connections.

        It uses a paging mechanism which defaults to 1024 connections.

        Arguments:
            sort: sorts the results. Default is connection ID.
            auth: include username. Default is False.
            subs: include subscriptions. Default is False. When set to "detail", a list with more detailed subscription information is returned.
            offset: pagination offset. Default is 0.
            limit: number of results to return. Default is 1024.
            cid: return result for a single connection by its id. Omitted by default.
            state: return results for connections of particular state. Default is "open".
            mqtt_client: return results for connections with this MQTT client id. Omitted by default.

        Returns:
            results as a dictionary.
        """
        if not isinstance(sort, SortOption):
            sort = SortOption(sort)
        if not isinstance(subs, SubsOption):
            subs = SubsOption(subs)
        if not isinstance(state, StateOption):
            state = StateOption(state)
        params = {
            "sort": sort.value,
            "auth": 1 if auth else 0,
            "subs": subs.value,
            "offset": offset,
            "limit": limit,
        }
        if cid:
            params["cid"] = int(cid)
        if mqtt_client:
            params["mqtt_client"] = mqtt_client
        return await self._request("/connz", **params)

    async def accountz(self, acc: t.Optional[str] = None) -> t.Dict[str, t.Any]:
        """The /accountz endpoint reports information on a server's active accounts.

        The default behavior is to return a list of all accounts known to the server.

        Arguments:
            acc: include metrics for the specified account only. Default is empty. When not set
                a list of all accounts is included.
        """
        params: t.Dict[str, t.Any] = {}
        if acc:
            params["acc"] = acc
        return await self._request("/accountz", **params)

    async def accstatz(self, unused: bool = False) -> t.Dict[str, t.Any]:
        """The /accstatz endpoint reports per-account statistics such as the number of connections, messages/bytes in/out, etc.

        Arguments:
            unused: include accounts that do not have any current connections when True. Default is False.

        Returns:
            results as a dictionary.
        """
        return await self._request("/accstatz", unused=unused)

    async def subsz(
        self,
        subs: bool = False,
        offset: int = 0,
        limit: int = 1024,
        test: t.Optional[str] = None,
    ) -> t.Dict[str, t.Any]:
        """The /subsz endpoint reports detailed information about the current subscriptions and the routing data structure.
        It is not normally used.

        Arguments:
            subs: include subscriptions. Default is false.
            offset: pagination offset. Default is 0.
            limit: number of results to return. Default is 1024.
            test: test whether a subscription exists.

        Returns:
            results as a dictionary.
        """
        params: t.Dict[str, t.Any] = {
            "subs": subs,
            "offset": offset,
            "limit": limit,
        }
        if test:
            params["test"] = test
        return await self._request("/subsz", **params)

    async def routez(
        self, subs: t.Union[bool, str, SubsOption] = SubsOption.FALSE
    ) -> t.Dict[str, t.Any]:
        """The /routez endpoint reports information on active routes for a cluster.

        Routes are expected to be low, so there is no paging mechanism with this endpoint.

        Arguments:
            subs: include subscriptions. Default is False. When set to "detail", a list with more details subscription information is returned.

        Returns:
            results as a dictionary.
        """
        if not isinstance(subs, SubsOption):
            subs = SubsOption(subs)
        return await self._request("/routez", subs=subs.value)

    async def leafz(self, subs: bool = False) -> t.Dict[str, t.Any]:
        """The /leafz endpoint reports detailed information about the leaf node connections.

        Arguments:
            subs: include internal subscriptions. Default is False.

        Returns:
            results as dict
        """
        return await self._request("/leafz", subs=subs)

    async def gatewayz(
        self,
        accs: bool = False,
        gw_name: t.Optional[str] = None,
        acc_name: t.Optional[str] = None,
    ) -> t.Dict[str, t.Any]:
        """The /gatewayz endpoint reports information about gateways used to create a NATS supercluster.

        Like routes, the number of gateways are expected to be low, so there is no paging mechanism with this endpoint.

        Arguments:
            accs: include account information. Default is false.
            gw_name: return only remote gateways with this name. Omitted by default.
            acc_name: limit the list of accounts to this account name. Omitted by default.

        Returns:
            results as dict
        """
        params: t.Dict[str, t.Any] = {"accs": bool(accs)}
        if gw_name:
            params["gw_name"] = gw_name
        if acc_name:
            params["acc_name"] = acc_name
        return await self._request("/gatewayz", **params)

    async def healthz(
        self,
        js_enabled: bool = False,
        js_server_only: bool = False,
    ) -> t.Dict[str, t.Any]:
        """The /healthz endpoint returns OK if the server is able to accept connections.

        Arguments:
            js_enabled: returns an error if jetstream is disabled. Omitted by default.
            js_server_only: skip healthcheck of accounts, streams and consumers. Omitted by default.

        Returns:
            results as dictionary.
        """
        params: t.Dict[str, t.Any] = {}
        if js_enabled:
            params["js_enabled"] = 1
        if js_server_only:
            params["js_server_only"] = 1
        return await self._request("/healthz", **params)

    async def close(self) -> None:
        if self._client and not self._client.is_closed:
            await self._client.aclose()

    async def __aenter__(self) -> "AsyncNATSMonitor":
        return self

    async def __aexit__(
        self,
        error_type: t.Optional[t.Type[BaseException]] = None,
        error: t.Optional[BaseException] = None,
        traceback: t.Optional[types.TracebackType] = None,
    ) -> None:
        await self.close()



import typing as t
from pathlib import Path

import jinja2


def load_template_from_path(template: t.Union[str, Path]) -> jinja2.Template:
    """Load a jinja2 template from path."""
    filepath = Path(template).absolute()
    if not filepath.exists():
        raise FileNotFoundError(filepath.as_posix())
    loader = jinja2.FileSystemLoader(filepath.parent)
    environment = jinja2.Environment(loader=loader)
    return environment.get_template(filepath.name)


def load_template_from_name(template: str) -> jinja2.Template:
    """Load a jinja2 template from name."""
    loader = jinja2.FileSystemLoader(Path(__file__).parent.joinpath("data"))
    environment = jinja2.Environment(loader=loader)
    return environment.get_template(template)



import typing as t
from pathlib import Path

import pytest
from _pytest.fixtures import SubRequest

from nats_tools.natsd import NATSD

F = t.TypeVar("F", bound=t.Callable[..., t.Any])


@pytest.fixture
def natsd(request: SubRequest) -> t.Iterator[NATSD]:
    if hasattr(request, "param"):
        params = dict(request.param)
    else:
        params = {"debug": True, "trace": True}
    if params.get("debug", None) is None:
        params["debug"] = True
    if params.get("trace", None) is None:
        params["trace"] = True
    with NATSD(**params) as daemon:
        yield daemon


def parametrize_nats_server(
    address: str = "127.0.0.1",
    port: int = 4222,
    client_advertise: t.Optional[str] = None,
    server_name: t.Optional[str] = None,
    server_tags: t.Optional[t.Dict[str, str]] = None,
    user: t.Optional[str] = None,
    password: t.Optional[str] = None,
    users: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
    token: t.Optional[str] = None,
    http_port: int = 8222,
    debug: t.Optional[bool] = None,
    trace: t.Optional[bool] = None,
    trace_verbose: t.Optional[bool] = None,
    logtime: t.Optional[bool] = None,
    pid_file: t.Union[str, Path, None] = None,
    port_file_dir: t.Union[str, Path, None] = None,
    log_file: t.Union[str, Path, None] = None,
    log_size_limit: t.Optional[int] = None,
    tls_cert: t.Union[str, Path, None] = None,
    tls_key: t.Union[str, Path, None] = None,
    tls_ca_cert: t.Union[str, Path, None] = None,
    cluster_name: t.Optional[str] = None,
    cluster_url: t.Optional[str] = None,
    cluster_listen: t.Optional[str] = None,
    routes: t.Optional[t.List[str]] = None,
    no_advertise: t.Optional[bool] = None,
    with_jetstream: bool = False,
    jetstream_domain: t.Optional[str] = None,
    store_directory: t.Union[str, Path, None] = None,
    max_memory_store: t.Optional[int] = None,
    max_file_store: t.Optional[int] = None,
    max_outstanding_catchup: t.Optional[int] = None,
    allow_leafnodes: bool = False,
    leafnodes_listen_address: t.Optional[str] = None,
    leafnodes_listen_port: t.Optional[int] = None,
    leafnode_remotes: t.Optional[t.Dict[str, t.Any]] = None,
    websocket_listen_address: t.Optional[str] = None,
    websocket_listen_port: t.Optional[int] = None,
    websocket_advertise_url: t.Optional[str] = None,
    websocket_tls: t.Optional[bool] = None,
    websocket_tls_cert: t.Union[str, Path, None] = None,
    websocket_tls_key: t.Union[str, Path, None] = None,
    websocket_same_origin: t.Optional[bool] = None,
    websocket_allowed_origins: t.Optional[t.List[str]] = None,
    websocket_compression: t.Optional[bool] = None,
    jwt_path: t.Union[str, Path, None] = None,
    operator: t.Optional[str] = None,
    system_account: t.Optional[str] = None,
    system_account_jwt: t.Optional[str] = None,
    allow_delete_jwt: t.Optional[bool] = None,
    compare_jwt_interval: t.Optional[str] = None,
    resolver_preload: t.Optional[t.Dict[str, str]] = None,
    config_file: t.Union[str, Path, None] = None,
    max_cpus: t.Optional[float] = None,
    start_timeout: float = 1,
) -> t.Callable[[F], F]:
    options = dict(
        address=address,
        port=port,
        client_advertise=client_advertise,
        server_name=server_name,
        server_tags=server_tags,
        user=user,
        password=password,
        users=users,
        token=token,
        http_port=http_port,
        debug=debug,
        trace=trace,
        trace_verbose=trace_verbose,
        logtime=logtime,
        pid_file=pid_file,
        port_file_dir=port_file_dir,
        log_file=log_file,
        log_size_limit=log_size_limit,
        tls_cert=tls_cert,
        tls_key=tls_key,
        tls_ca_cert=tls_ca_cert,
        cluster_name=cluster_name,
        cluster_url=cluster_url,
        cluster_listen=cluster_listen,
        routes=routes,
        no_advertise=no_advertise,
        with_jetstream=with_jetstream,
        jetstream_domain=jetstream_domain,
        store_directory=store_directory,
        max_memory_store=max_memory_store,
        max_file_store=max_file_store,
        max_outstanding_catchup=max_outstanding_catchup,
        allow_leafnodes=allow_leafnodes,
        leafnodes_listen_address=leafnodes_listen_address,
        leafnodes_listen_port=leafnodes_listen_port,
        leafnode_remotes=leafnode_remotes,
        websocket_listen_address=websocket_listen_address,
        websocket_listen_port=websocket_listen_port,
        websocket_advertise_url=websocket_advertise_url,
        websocket_tls=websocket_tls,
        websocket_tls_cert=websocket_tls_cert,
        websocket_tls_key=websocket_tls_key,
        websocket_same_origin=websocket_same_origin,
        websocket_allowed_origins=websocket_allowed_origins,
        websocket_compression=websocket_compression,
        jwt_path=jwt_path,
        operator=operator,
        system_account=system_account,
        system_account_jwt=system_account_jwt,
        allow_delete_jwt=allow_delete_jwt,
        compare_jwt_interval=compare_jwt_interval,
        resolver_preload=resolver_preload,
        config_file=config_file,
        start_timeout=start_timeout,
        max_cpus=max_cpus,
    )
    return pytest.mark.parametrize("natsd", [options], indirect=True)



import typing as t
from pathlib import Path

from .utils import load_template_from_name, load_template_from_path


class ConfigGenerator:
    def __init__(self, template: t.Union[str, Path] = "default.conf.j2") -> None:
        """Create a new instance of config generator.

        Arguments:
            template: the template used to render configuration.
        """
        if isinstance(template, Path):
            self.template = load_template_from_path(template)
        elif Path(template).is_file():
            self.template = load_template_from_path(template)
        else:
            self.template = load_template_from_name(template)

    def render(
        self,
        address: str = "127.0.0.1",
        port: int = 4222,
        client_advertise: t.Optional[str] = None,
        server_name: t.Optional[str] = None,
        server_tags: t.Optional[t.Dict[str, str]] = None,
        user: t.Optional[str] = None,
        password: t.Optional[str] = None,
        users: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
        token: t.Optional[str] = None,
        http_port: int = 8222,
        debug: t.Optional[bool] = None,
        trace: t.Optional[bool] = None,
        trace_verbose: t.Optional[bool] = None,
        logtime: t.Optional[bool] = None,
        pid_file: t.Union[str, Path, None] = None,
        port_file_dir: t.Union[str, Path, None] = None,
        log_file: t.Union[str, Path, None] = None,
        log_size_limit: t.Optional[int] = None,
        tls_cert: t.Union[str, Path, None] = None,
        tls_key: t.Union[str, Path, None] = None,
        tls_ca_cert: t.Union[str, Path, None] = None,
        cluster_name: t.Optional[str] = None,
        cluster_url: t.Optional[str] = None,
        cluster_listen: t.Optional[str] = None,
        routes: t.Optional[t.List[str]] = None,
        no_advertise: t.Optional[bool] = None,
        with_jetstream: bool = False,
        jetstream_domain: t.Optional[str] = None,
        store_directory: t.Union[str, Path, None] = None,
        max_memory_store: t.Optional[int] = None,
        max_file_store: t.Optional[int] = None,
        max_outstanding_catchup: t.Optional[int] = None,
        allow_leafnodes: bool = False,
        leafnodes_listen_address: t.Optional[str] = None,
        leafnodes_listen_port: t.Optional[int] = None,
        leafnode_remotes: t.Optional[t.Dict[str, t.Any]] = None,
        websocket_listen_address: t.Optional[str] = None,
        websocket_listen_port: t.Optional[int] = None,
        websocket_advertise_url: t.Optional[str] = None,
        websocket_tls: t.Optional[bool] = None,
        websocket_tls_cert: t.Union[str, Path, None] = None,
        websocket_tls_key: t.Union[str, Path, None] = None,
        websocket_same_origin: t.Optional[bool] = None,
        websocket_allowed_origins: t.Optional[t.List[str]] = None,
        websocket_compression: t.Optional[bool] = None,
        jwt_path: t.Union[str, Path, None] = None,
        operator: t.Optional[str] = None,
        system_account: t.Optional[str] = None,
        system_account_jwt: t.Optional[str] = None,
        allow_delete_jwt: t.Optional[bool] = None,
        compare_jwt_interval: t.Optional[str] = None,
        resolver_preload: t.Optional[t.Dict[str, str]] = None,
    ) -> str:
        """Render configuration according to arguments."""
        kwargs: t.Dict[str, t.Any] = {}

        kwargs["server_host"] = address
        kwargs["server_port"] = port
        kwargs["client_advertise"] = client_advertise
        kwargs["server_name"] = server_name
        kwargs["http_port"] = http_port

        if debug is not None:
            kwargs["debug"] = debug
        if trace is not None:
            kwargs["trace"] = trace
        if trace_verbose is not None:
            kwargs["trace_verbose"] = trace_verbose
        if logtime is not None:
            kwargs["logtime"] = logtime
        if pid_file is not None:
            kwargs["pid_file"] = Path(pid_file).as_posix()
        if port_file_dir is not None:
            kwargs["port_file_dir"] = Path(port_file_dir).as_posix()
        if log_file is not None:
            kwargs["log_file"] = Path(log_file).as_posix()
        if log_size_limit is not None:
            kwargs["log_size_limit"] = log_size_limit
        if server_tags:
            kwargs["server_tags"] = [
                f"{key}:{value}" for key, value in server_tags.items()
            ]

        cluster = False
        if cluster_listen or cluster_url:
            if cluster_listen is None:
                cluster_listen = cluster_url
            cluster = True
            kwargs["cluster_listen"] = cluster_listen
            if cluster_url is not None:
                kwargs["cluster_url"] = cluster_url
            if routes is not None:
                kwargs["routes"] = routes
            if no_advertise is not None:
                kwargs["no_advertise"] = no_advertise
            if cluster_name is not None:
                kwargs["cluster_name"] = cluster_name
        kwargs["cluster"] = cluster

        tls = False
        if tls_cert or tls_key:
            if not (tls_cert and tls_key):
                raise ValueError(
                    "tls_cert and tls_key argument must be provided together"
                )
            tls = True
            tls_cert_file = Path(tls_cert).as_posix()
            tls_key_file = Path(tls_key).as_posix()
            kwargs["tls_cert_file"] = tls_cert_file
            kwargs["tls_key_file"] = tls_key_file
            if tls_ca_cert:
                tls_ca_file = Path(tls_ca_cert).as_posix()
                kwargs["tls_ca_file"] = tls_ca_file
        kwargs["tls"] = tls
        kwargs["enable_jetstream"] = with_jetstream
        kwargs["jetstream_domain"] = jetstream_domain
        kwargs["max_file_store"] = max_file_store
        kwargs["max_memory_store"] = max_memory_store
        kwargs["max_outstanding_catchup"] = max_outstanding_catchup
        if store_directory is not None:
            kwargs["jetstream_store_dir"] = store_directory

        if user or password:
            if not (user and password):
                raise ValueError(
                    "Both user and password argument must be provided together"
                )

        if token:
            if user:
                raise ValueError(
                    "token argument cannot be used together with user and password"
                )

        if users:
            if token or user:
                raise ValueError(
                    "users argument cannot be used with token or user and password"
                )

        if operator:
            if users or token or user:
                raise ValueError(
                    "operator argument cannot be used with any of users, token, user and password arguments"
                )
            if system_account is None:
                raise ValueError("system_account argument must be provided")
            if system_account_jwt is None:
                raise ValueError("system_account_jwt argument must be provided")
            if jwt_path is None:
                raise ValueError("jwt_path argument must be provided")

        kwargs["user"] = user
        kwargs["password"] = password
        kwargs["users"] = users
        kwargs["token"] = token

        kwargs["operator"] = operator
        kwargs["system_account"] = system_account
        kwargs["jwt_path"] = jwt_path
        jwts = resolver_preload or {}
        if system_account and system_account_jwt:
            jwts[system_account] = system_account_jwt
        kwargs["jwts"] = jwts
        kwargs["allow_delete_jwt"] = allow_delete_jwt or False
        kwargs["compare_jwt_interval"] = compare_jwt_interval or "2m"

        if leafnodes_listen_address or leafnodes_listen_port:
            leafnodes_listen_address = leafnodes_listen_address or address
            leafnodes_listen_port = leafnodes_listen_port or 7422
            allow_leafnodes = True
            kwargs["leafnodes_listen_address"] = leafnodes_listen_address
            kwargs["leafnodes_listen_port"] = leafnodes_listen_port
        kwargs["allow_leafnodes"] = allow_leafnodes
        kwargs["leafnode_remotes"] = leafnode_remotes

        websocket = False
        if websocket_listen_port or websocket_listen_address:
            if websocket_listen_address is None:
                websocket_listen_address = address
            if websocket_listen_port is None:
                if websocket_tls or websocket_tls_cert:
                    websocket_listen_port = 443
            websocket = True
            kwargs["websocket_listen_port"] = websocket_listen_port
            kwargs["websocket_listen_address"] = websocket_listen_address
            if websocket_advertise_url:
                kwargs["websocket_advertise_url"] = websocket_advertise_url
            if websocket_tls_cert and websocket_tls_key:
                if not websocket_tls_cert and websocket_tls_key:
                    raise ValueError(
                        "websocket_tls_cert and websocket_tls_key must be provided to enable websocket TLS"
                    )
            if (
                (tls and websocket_tls) or (tls and websocket_tls is None)
            ) and websocket_tls_cert is None:
                if tls_cert is None or tls_key is None:
                    raise ValueError(
                        "websocket_tls_cert and websocket_tls_key must be provided to enable websocket TLS"
                    )
                websocket_tls_cert = Path(tls_cert).as_posix()
                websocket_tls_key = Path(tls_key).as_posix()
            websocket_tls = False
            if websocket_tls_cert:
                websocket_tls = True
                kwargs["websocket_tls_cert_file"] = websocket_tls_cert
                kwargs["websocket_tls_key_file"] = websocket_tls_key
            kwargs["websocket_tls"] = websocket_tls
            if websocket_tls:
                if websocket_same_origin is not None:
                    kwargs["websocket_same_origin"] = websocket_same_origin
                kwargs["websocket_allowed_origins"] = websocket_allowed_origins
            if websocket_compression is not None:
                kwargs["websocket_compression"] = websocket_compression
        kwargs["websocket"] = websocket

        return self.template.render(**kwargs)
